#!/usr/bin/env python

# A script for handling invocation schema generation and validation for Boutiques application descriptors
# Requires jsonschema 2.5

import json
import jsonschema
import os
import sys
import argparse
from functools import reduce
from boutiques.validator import validate_descriptor

# Generate an invocation schema from a Boutiques application descriptor
def generateInvocationSchema(toolDesc, oname=None, validateWrtMetaSchema=True):
  # Subclass dictionary to act more like a ruby hash map
  class RMap(dict):
    def __getitem__(self, key): return dict.get(self, key)
  s = RMap(toolDesc)
  descName, inputs, groups = s['name'], s['inputs'], s['groups'] or []
  oname = descName + ".invocationSchema" if oname is None else oname
  schema = {
    "$schema"              : "http://json-schema.org/draft-04/schema#",
    "title"                : str(oname),
    "description"          : "Invocation schema for " + str(descName.lower()) + ".",
    "type"                 : "object",
    "additionalProperties" : False
  }
  if inputs is None: return schema
  # Handle type constraints for single inputs
  def addTypeConstraints(h,ind):
    i = RMap( ind )
    id, type, target = i['id'], i['type'], None
    def undertype():
      if i['integer']: return { 'type' : 'integer' }
      elif i['value-choices']: return { 'enum' : i['value-choices'] }
      elif type == "File": return { 'type' : 'string' }
      elif type == "Flag": return { 'type' : 'boolean' }
      else: return { 'type' : type.lower() }
    if i['list']:
      h[id] = { 'type'  : 'array', 'items' : undertype() }
      if i["min-list-entries"]: h[id]["minItems"] = i["min-list-entries"]
      if i["max-list-entries"]: h[id]["maxItems"] = i["max-list-entries"]
      target = h[id]["items"]
    else:
      h[id] = undertype()
      target = h[id]
    if not i["minimum"] is None: target["minimum"] = i["minimum"] # Python treats 0 as false
    if not i["maximum"] is None: target["maximum"] = i["maximum"]
    if i["exclusive-minimum"]: target["exclusiveMinimum"] = i["exclusive-minimum"]
    if i["exclusive-maximum"]: target["exclusiveMaximum"] = i["exclusive-maximum"]
    return h
  schema["properties"] = reduce(addTypeConstraints, inputs, {})
  # Required inputs
  reqInputs = [x['id'] for x in [x for x in inputs if not x.get('optional')]]
  if len(list(reqInputs)) > 0: schema["required"] = reqInputs
  # Helper functions (assumes properly formed descriptor)
  byInd  = lambda id: [i for i in inputs if id==i['id']][0]
  isFlag = lambda id: (byInd(id)['type'] == 'Flag')
  # Handle one-is-required groups
  g1r = [g for g in groups if g.get("one-is-required")]
  def reqMember(m): # Generate code for a one-is-required member
    if isFlag(m): return { "properties" : { m : { "enum" : [ True ] } }, "required" : [m] }
    else: return { "required" : [m] }
  if len(g1r) > 0: schema["allOf"] = [{"anyOf" : list(map(reqMember, g['members']))} for g in g1r]
  # Handle requires and disables-inputs/mutex constraints
  def handleDisablesRequires(h,inval):
    i, h = RMap( inval ), RMap( h )
    id, reqs, disbs = i['id'], i['requires-inputs'], i['disables-inputs'] or []
    # Mutex group members added to disablees list
    for g in [g for g in groups if g.get('mutually-exclusive') and (id in g['members'])]:
      for m in [m for m in g['members'] if not m==id]: disbs.append( m )
    # Handle requires-inputs (flags have to be true to satisfy it)
    if reqs:
      reqMap = { r : { "enum" : [ True ] } for r in reqs if isFlag(r) }
      if isFlag( id ):
        h[id] = {}
        h[id]['anyOf'] = [
          { "properties" : reqMap, "required" : reqs },
          { "properties" : { id : { "enum" : [ False ] } } }
        ]
      else:
        h[id] = { "required" : reqs }
        h[id]['properties'] = reqMap
    # Handle disables-inputs (false flags do not violate the disabled criteria)
    if len(disbs) > 0:
      if h[id] is None: h[id] = RMap( {} )
      def makeSingleDisablesMap(dh, d): # Flags can be non-present or false
        if isFlag( d ): dh[d] = { "oneOf" : [ { "not" : {} }, { "enum" : [ False ] } ] }
        else: dh[d] = { "not" : {} }
        return dh
      disbMap = reduce(makeSingleDisablesMap, disbs, {})
      if isFlag( id ): # Flags do not disable their targets when false
        if h[id].get('anyOf') is None:
          h[id]['anyOf'] = [
            { "properties" : disbMap },
            { "properties" : { id : { "enum" : [ False ] } } }
          ]
        else: # Need to add disables properties to requires ones, if the latter exist
          h[id]['anyOf'][0]['properties'].update( disbMap )
      else:
        if h[id].get('properties') is None: h[id]['properties'] = RMap( {} )
        h[id]['properties'].update( disbMap )
    return h
  schema["dependencies"] = reduce(handleDisablesRequires, inputs, {})
  # Validate wrt meta-schema if desired
  if validateWrtMetaSchema: validateSchema( schema )
  # Return final schema
  return schema

# Validate data with respect to the invocation schema
def validateSchema(s, d=None):
  # Check schema wrt meta-schema
  try:
    jsonschema.Draft4Validator.check_schema(s)
  except jsonschema.SchemaError as se:
    errExit("Invocation schema is invalid.\n" + str(se.message), False)
  # Check data instance against schema
  if d:
    try:
      jsonschema.validate(d, s)
    except jsonschema.ValidationError as e:
      raise jsonschema.ValidationError(e.message)
    print("Invocation Schema validation OK")
          
# Script exit helper
def errExit(msg, parser, err_code = 1):
    sys.stderr.write('Error: ' + msg + '\n')
    sys.exit( err_code )




#!/usr/bin/env python

# Copyright 2015 - 2017:
#   The Royal Institution for the Advancement of Learning McGill University,
#   Centre National de la Recherche Scientifique,
#   University of Southern California,
#   Concordia University
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter
from jsonschema import ValidationError
from boutiques.validator import validate_descriptor
from git import Repo
from github import Github
import git, json, os, sys
try:
    # For Python 3.0 and later
    from urllib.request import urlopen, URLError
except ImportError:
    # Fall back to Python 2's urllib2
    from urllib2 import urlopen, URLError

class Publisher():

    def __init__(self, boutiques_dir, remote_name, tool_author, tool_url, inter,
                 neurolinks_repo_url, neurolinks_dest_path, github_user, github_password, no_github):
        self.boutiques_dir = os.path.abspath(boutiques_dir)
        self.remote_name = remote_name if remote_name else 'origin'
        try:
            self.boutiques_repo = Repo(self.boutiques_dir)
        except git.exc.InvalidGitRepositoryError as e:
            # We won't publish a descriptor which is not in a Git repo!
            sys.stderr.write('error: %s is not a Git repository.\n' % self.boutiques_dir)
            sys.exit(1)

        for remote in self.boutiques_repo.remotes:
            if remote.name == self.remote_name:
                self.boutiques_remote = remote

        url = self.boutiques_remote.url
        # If remote URL is on Github, try to guess the URL of the Boutiques descriptor
        if "github.com" in url: 
            url = url.replace(".git","").replace("git@github.com:","https://github.com/")
            url = url.replace("https://github.com/", "https://raw.githubusercontent.com/")
            url = url.replace("http://github.com/", "https://raw.githubusercontent.com/")
            url += "master"
        self.base_url = url

        # Try to guess the tool author
        self.tool_author = tool_author

        self.tool_url = tool_url if tool_url else "http://example.com"
        self.inter = inter
        self.no_github = no_github

        if self.no_github:
            return

        # Set github credentials
        self.github_user = github_user
        self.github_password = github_password
        self.fix_github_credentials()
        # Fork and clone the neurolinks repo
        if neurolinks_repo_url.startswith("http"):
            fork_url = self.fork_repo(neurolinks_repo_url) # won't do anything if fork already exists
            neurolinks_local_url = self.clone_repo(fork_url, neurolinks_dest_path)
            # Add base neurolinks repo as remote and pull from it, in case
            # fork already existed and is outdated
            self.neurolinks_repo = Repo(neurolinks_local_url)
            base = self.neurolinks_repo.create_remote('base', neurolinks_repo_url)
            neurolinks_repo_url = neurolinks_local_url
        else:
            self.neurolinks_repo = Repo(neurolinks_repo_url)
        self.neurolinks_repo.remotes.base.pull('master')
        
        self.tools_file = os.path.abspath(os.path.join(neurolinks_repo_url,
                                                       'jsons', 'tool.json'))
        if not os.path.isfile(self.tools_file):
            print("error: cannot find {0} (check neurolinks repo).".format(self.tools_file))
            sys.exit(1)
            
    def fix_github_credentials(self):
        pygithub_file = os.path.join(os.getenv("HOME"),".pygithub")
        # Read saved credentials
        try:
            with open(pygithub_file,"r") as f:
                json_creds = json.load(f)
        except IOError:
            json_creds = {}
        except ValueError:
            json_creds = {}
        # Credentials passed as parameters have precedence. Ask credentials if none
        # can be found
        username = self.github_user if self.github_user else json_creds.get('user')
        if not username:
            username = self.get_from_stdin("GitHub login")
        password = self.github_password if self.github_password else json_creds.get('password')
        if not password:
            password = self.get_from_stdin("GitHub password (user: {0})".format(username))
        # Save credentials
        self.github_username = username
        self.github_password = password
        json_creds['user'] = self.github_username
        json_creds['password'] = self.github_password
        with open(pygithub_file,"w") as f:
            f.write(json.dumps(json_creds, indent=4, sort_keys=True))
        
    def find_json_files(self, directory):
        json_files = []
        for root, dirs, files in os.walk(directory):
            for file in files:
                if file.endswith(".json"):
                    json_files.append(os.path.join(root, file))
        return json_files
    
    def get_json_string(self, identifier, label, description, tool_author, tool_url, boutiques_url, docker_container, singularity_container):

        path, fil = os.path.split(__file__)
        template_file = os.path.join(path, "neurolinks-template", "tool.json")

        with open(template_file) as f:
            template_string = f.read()

        template_string = template_string.replace("@@__ID__@@", identifier)
        template_string = template_string.replace("@@__LABEL__@@", label)
        template_string = template_string.replace("@@__DESCRIPTION__@@", description)
        template_string = template_string.replace("@@__TOOL_AUTHOR__@@", tool_author)
        template_string = template_string.replace("@@__TOOL_URL__@@", tool_url)
        template_string = template_string.replace("@@__BOUTIQUES_URL__@@", boutiques_url)

        json_string = json.loads(template_string)
        if docker_container:
            json_string['dockercontainer'] = docker_container
        if singularity_container:
            json_string['singularitycontainer'] = singularity_container
            
        return json_string

    def is_boutiques_descriptor(self, json_file):
        try:
            validate_descriptor(json_file)
            return True
        except:
            return False

    def get_from_stdin(self, question, default_value=None, input_type=None):
        if not self.inter:
            return default_value
        prompt = question+" ("+default_value+"): " if default_value else question+": "
        try:
            answer = raw_input(prompt) # Python 2
        except NameError:
            answer = input(prompt) # Python 3
        answer = answer if answer else default_value
        if input_type == "URL":
            while not self.check_url(answer):
                answer = self.get_from_stdin(question, default_value, input_type)
        return answer

    def check_url(self, url):
        try:
            code = urlopen(url).getcode()
        except ValueError as e:
            print("error: {0}".format(e.message))
            return False
        except URLError as e:
            print("error: {0}".format(e.message))
            return False
        if code != 200:
            print("warning: URL is not accessible (status: {0})".format(code))
            return False
        else:
            print("  ... URL is accessible.")
            return True
            
    
    def print_banner(self, descriptor):
        name = descriptor['name']
        text = "Found Boutiques tool {0}".format(name,"")
        line = ""
        for i in range(0, len(text)):
            line = line + "="
        print("")
        print(line)
        print(text)
        print(line)
        
    def get_descriptors(self):
        json_strings = []
        json_files = self.find_json_files(self.boutiques_dir)
        descriptors = [ x for x in json_files if self.is_boutiques_descriptor(x) ]
        return descriptors

    def get_neurolinks_json(self, descriptor_file_name, tools):
        entities = tools.get('entities')
        with open(descriptor_file_name, "r") as f:
            descriptor = json.load(f)
        self.print_banner(descriptor)
        if self.inter:
            if(self.get_from_stdin("Publish Y/n?","Y") != "Y"):
                return None
        if(self.contains(descriptor['name'],entities)):
            if not self.inter:
                print('Not overwriting existing entry - use interactive mode to override')
                return None
            if(self.get_from_stdin("{0} is already published, overwrite Y/n?".format(
                    descriptor['name']),"n") != "Y"):
                return None
            tools['entities'] = self.remove(descriptor['name'], entities)
        label = descriptor['name']
        description = descriptor.get('description')
        docker_container = None
        singularity_container = None
        container_image = descriptor.get('container-image')
        if container_image:
            if container_image.get('type') == "docker":
                index = container_image.get('index') if container_image.get('index') else 'http://index.docker.io'
                if "index.docker.io" in index:
                    index = "https://hub.docker.com/r/"
                elif "quay.io" in index:
                    index = "https://quay.io/repository"
                docker_container = os.path.join(index,container_image.get("image").split(':')[0])
            if container_image.get('type') == "singularity":
                index = container_image.get('index') if container_image.get('index') else 'shub://'
                if index == "docker://":
                    singularity_container = os.path.join("https://hub.docker.com",container_image.get("image").split(':')[0]) 
                else:
                    singularity_container = os.path.join(index,container_image.get("image")) 
        identifier = label.replace(" ","_")
        boutiques_url = self.get_url(descriptor_file_name)
        if self.inter:
            self.tool_author = self.get_from_stdin("Tool author",
                                                   self.tool_author)
            self.tool_url = self.get_from_stdin("Tool URL", self.tool_url, "URL")
            boutiques_url = self.get_from_stdin("Boutiques descriptor URL",
                                                boutiques_url, "URL")
        return self.get_json_string(identifier, label, description,
                                                 self.tool_author, self.tool_url,
                                                 boutiques_url, docker_container, singularity_container)
    def get_url(self, file_path):
        file_path = file_path.replace(self.boutiques_dir,"")
        url = self.base_url + file_path
        return url

    def contains(self, label, tools):
        if not tools:
            return
        for tool in tools:
            if tool['label'] == label:
                return True
        return False

    def remove(self, label, tools):
        new_tools = [ x for x in tools if x['label'] != label ]
        return new_tools
    
    def publish(self):
        json_tools = []
        existing_tools = {}
        if not self.no_github:
            # Load existing tools
            with open(self.tools_file, "r") as json_file:
                existing_tools = json.load(json_file)
        # Get new tools from boutiques repo
        descriptors = self.get_descriptors()
        for descriptor_file_name in descriptors:
            # Build neurolinks string
            neurolinks_json = self.get_neurolinks_json(descriptor_file_name, existing_tools)
            # Have user review before commit
            if not self.inter and neurolinks_json:
                print("Tool summary:")
                print(json.dumps(neurolinks_json, indent=4, sort_keys=True))
                if(self.get_from_stdin("Publish Y/n?","n") != "Y"):
                    neurolinks_json = None
            # Publish json string unless user didn't want to
            if neurolinks_json:
                json_tools.append(neurolinks_json)
        if len(json_tools) == 0:
            print("Nothing to publish, bye!")
            return
        if self.no_github:
            for tool in json_tools:
                print(json.dumps(tool, indent=4, sort_keys=True))
            return
        # Add new tools to existing tools
        for tool in json_tools:
            existing_tools['entities'].append(tool)
        # Write updated tools
        json_file = open(self.tools_file, "w")
        json_file.write(json.dumps(existing_tools, indent=4, sort_keys=True))
        json_file.close()
        # Commit tools file
        self.neurolinks_repo.index.add([self.tools_file])
        self.neurolinks_repo.index.commit("Added tool {0} with bosh-publish".format(tool['label']))
        # Push to fork
        self.neurolinks_repo.remotes.origin.push()
        # Make PR
        self.pr(self.neurolinks_repo.remotes.origin.url,
                self.neurolinks_repo.remotes.base.url)
            
    def clone_repo(self, repo_url, dest_path):
        print("Cloning {0} to {1}...".format(repo_url,dest_path))
        Repo.clone_from(repo_url, dest_path)
        return os.path.abspath(dest_path)

    def fork_repo(self, repo_url):
        print("Forking {0} to {1}...".format(repo_url, self.github_username))
        g = Github(self.github_username, self.github_password)
        github_user = g.get_user()
        repo = g.get_repo("brainhack101/neurolinks")
        myfork = github_user.create_fork(repo)
        return myfork.ssh_url        

    def pr(self, fork_url, base_url):
        print("Create a pull request from {0} to {1} to finalize publication. This cannot be done automatically.".format(fork_url,base_url))




#!/usr/bin/env python

import argparse, os, sys, json, random as rnd, string, math, random, subprocess, time, pwd

# Executor class
class LocalExecutor(object):
  """
  This class represents a json descriptor of a tool, and can execute various tasks related to it.
  It is constructed first via an input json descriptor file, which is held in the desc_dict field.

  An input can be added to it via the in_dict field, a dictionary from param ids to values.
  The in_dict field should only be modified via the public readInput method, which can either take
  a file (json or csv) or a string written in the command line. The field is always validated by checking
  the input parameters with respect to the descriptor.

  Other public methods include:
    execute - attempts to execute the tool described by the descriptor based on the current input (in in_dict)
    printCmdLine - simply prints the generated command line based on the current input values
    generateRandomParams - fills in_dict with random values (following the constraints of the descriptor schema)
  """

  # Constructor
  def __init__(self, desc, options={}):
    ## Initial parameters ##
    self.desc_path = desc  # Save descriptor path
    self.errs      = []    # Empty errors holder
    self.debug     = False # debug mode (also use python -u for unbuffered printing)
    ## Parse JSON descriptor ##
    with open(desc, 'r') as descriptor:
      self.desc_dict = json.loads( descriptor.read() )
    ## Helpers Functions ##
    # The set of input parameters from the json descriptor
    self.inputs     = self.desc_dict['inputs'] # Structure: [ {id:...},...,{id:...} ]
    # The set of output parameters from the json descriptor
    self.outputs     = self.desc_dict['output-files'] # Structure: [ {id:...},...,{id:...} ]
    # The set of parameter groups, according to the json descriptor
    self.groups     = self.desc_dict['groups'] if 'groups' in list(self.desc_dict.keys()) else []
    # Retrieves the parameter corresponding to the given id
    self.byId       = lambda n: [ v for v in self.inputs+self.outputs if v['id']==n][0]
    # Retrieves the group corresponding to the given id
    self.byGid      = lambda g: [ v for v in self.groups if v['id']==g][0]
    # Retrieves the value of a field of an input from the descriptor. Returns None if not present.
    self.safeGet    = lambda i, k: None if not k in list(self.byId(i).keys()) else self.byId(i)[k]
    # Retrieves the value of a field of a group from the descriptor. Returns None if not present.
    self.safeGrpGet = lambda g, k: None if not k in list(self.byGid(g).keys()) else self.byGid(g)[k]
    # Retrieves the group a given parameter id belongs to; otherwise, returns None
    self.assocGrp   = lambda i: ([g for g in self.groups if i in g["members"]] or [None])[0]
    # Returns the required inputs of a given input id, or the empty string
    self.reqsOf     = lambda t: self.safeGet(t,"requires-inputs") or []
    ## Container-image Options ##
    self.con = self.desc_dict.get('container-image')
    self.launchDir = None if self.con is None else self.con.get('working-directory')
    ## Extra Options ##
    # Include: forcePathType and destroyTempScripts
    for option in list(options.keys()): setattr(self, option, options.get(option))
    # Container Implementation check
    conEngines = ['docker', 'singularity']
    if (not self.con is None) and self.con['type'] not in conEngines:
        msg = 'Other container types than {} (e.g. {}) are not yet supported'
        raise ValueError(msg.format(", ".join(conEngines), self.con['type']))

  # Attempt local execution of the command line generated from the input values
  def execute(self, mount_strings):
    '''
    The execute method runs the generated command line (from either generateRandomParams or readInput).
    If docker is specified, it will attempt to use it, instead of local execution.
    After execution, it checks for output file existence.
    '''    
    command, exit_code, con = self.cmdLine[0], None, self.con or {}
    print('Attempting execution of command:\n\t' + command + '\n---/* Start program output */---')
    # Check for Container image
    conType, conImage = con.get('type'), con.get('image')
    conIsPresent = (not conImage is None) 
    # Export environment variables, if they are specified in the descriptor
    envVars = {}
    if 'environment-variables' in list(self.desc_dict.keys()):
      for (envVarName,envVarValue) in [ (p['name'],p['value']) for p in self.desc_dict['environment-variables'] ]:
        os.environ[envVarName], envVars[envVarName] = envVarValue, envVarValue # for non-container and docker resp.
    # Container script constant name
    # Note that docker/singularity cannot do a local volume mount of files starting with a '.', hence this one does not
    millitime = int(time.time()*1000)
    dsname = 'temp-' + str(random.SystemRandom().randint(0,int(millitime))) + "-" + str(millitime) + '.localExec.boshjob.sh' # time tag to avoid overwrites
    # If container is present, alter the command template accordingly
    if conIsPresent:
      if conType == 'docker':
        # Pull the docker image
        if self._localExecute("docker pull " + str(conImage)):
          print("Container not found online - trying local copy")
      elif conType == 'singularity':
        # Pull the docker image
        if self._localExecute("singularity pull shub://" + str(conImage)):
          print("Container not found online - trying local copy")
      else:
        print('Unrecognized container type: \"%s\"'%conType) 
        sys.exit(1)
      # Generate command script
      uname, uid = pwd.getpwuid( os.getuid() )[ 0 ], str(os.getuid())
      # Adds the user to the container before executing the templated command line
      userchange = '' if not self.changeUser else ("useradd --uid " + uid + ' ' + uname + "\n")
      # If --changeUser was desired, run with su so that any output files are owned by the user instead of root
      # Get the supported shell by the docker or singularity
      if self.changeUser: command = 'su ' + uname + ' -c ' + "\"{0}\"".format(command)
      cmdString = "#!/bin/sh -l\n" + userchange + str( command )
      with open(dsname,"w") as scrFile: scrFile.write(cmdString)
      # Ensure the script is executable
      self._localExecute( "chmod 755 " + dsname )
      # Prepare extra environment variables
      envString = " "
      if envVars:
        for (key,val) in list(envVars.items()): envString += "-e " + str(key) + "=\'" + str(val) + '\' '
      # Change launch (working) directory if desired
      launchDir = '${PWD}' if (self.launchDir is None) else self.launchDir
      # Run it in docker
      mount_strings = [] if not mount_strings else mount_strings
      mount_strings.append('${PWD}:' + launchDir)
      if conType == 'docker':
        # export mounts to docker string
        docker_mounts = " -v ".join(m for m in mount_strings)
        dcmd = 'docker run --entrypoint=/bin/sh --rm' + envString + ' -v '+docker_mounts+ ' -w ' + launchDir + ' ' + str(conImage) + ' ./' + dsname
      elif conType == 'singularity':
        singularity_mounts = " -B ".join(m for m in mount_strings)
        #TODO: Test singularity runtime on cluster
        dcmd = 'singularity exec' + envString + ' -B ' + singularity_mounts + ' -W ' + launchDir + ' ' + str(conImage) + ' ./' + dsname
      else:
        print('Unrecognized container type: \"%s\"'%conType)
        sys.exit(1)
      print('Executing via: ' + dcmd)
      (stdout, stderr), exit_code = self._localExecute( dcmd )
    # Otherwise, just run command locally
    else:
      (stdout, stderr), exit_code = self._localExecute( command )
    if stdout != '': print('Execution output: '+ str(stdout))
    # Report exit status
    print('---/* End program output */---\nCompleted execution (exit code: ' + str(exit_code) + ')')
    time.sleep(0.5) # Give the OS a (half) second to finish writing
    # Destroy temporary docker script, if desired. By default, keep the script so the dev can look at it.
    if conIsPresent and self.destroyTempScripts:
      if os.path.isfile(dsname): os.remove(dsname)
    # Check for output files (note: the path-template can contain value-keys)
    print('Looking for output files:')
    for outfile in self.desc_dict['output-files']:
      outFileName = self.out_dict[outfile['id']]
      # Look for the target file
      exists = os.path.exists( outFileName )
      # Note whether it could be found or not
      isOptional = outfile['optional'] if 'optional' in list(outfile.keys()) else False
      s1 = 'Optional' if isOptional else 'Required'
      s2 = '' if exists else 'not '
      err = "Error! " if (not isOptional and not exists) else '' # Add error warning when required file is missing
      print("\t"+err+s1+" output file \'"+outfile['name']+"\' was "+s2+"found at "+ outFileName)
    desc_err = ''
    if 'error-codes' in list(self.desc_dict.keys()):
      for err_elem in self.desc_dict['error-codes']:
         if err_elem['code'] == exit_code:
            desc_err = err_elem['description']
            break
    error_msg = ''
    if stderr != '':
        error_msg = 'Execution ERR ({0}): {1}'.format(exit_code, stderr)
    if desc_err != '':
        error_msg += '{0}{1} ERR ({2}): {3}'.format('\n' if error_msg != '' else '', self.desc_dict['name'], exit_code, desc_err)
    
    return exit_code, stdout, error_msg

  # Private method that attempts to locally execute the given command. Returns the exit code.
  def _localExecute(self,command):
    try: # Note: invokes the command through the shell (potential injection dangers)
      process = subprocess.Popen(command, shell=True,
                                 stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE)

    except OSError as e:
      sys.stderr.write('OS Error during attempted execution!')
      raise e
    except ValueError as e:
      sys.stderr.write( 'Input Value Error during attempted execution!' )
      raise e
    else:
      return process.communicate(), process.returncode


  # Private method to generate a random input parameter set that follows the constraints from the json descriptor
  # This method fills in the in_dict field of the object with constrained random values
  def _randomFillInDict(self):
    ## Private helper functions for filling the dictionary ##
    # Helpers for generating random numbers, strings, etc...
    # Note: uses-absolute-path is satisfied for files by the automatic replacement in the _validateDict
    nd, nl   = 2, 5 # nd = number of random characters to use in generating strings, nl = max number of random list items
    randDigs = lambda: ''.join(rnd.choice(string.digits) for _ in range(nd)) # Generate random string of digits
    randFile = lambda: 'f_' + randDigs() + rnd.choice(['.csv','.tex','.j','.cpp','.m','.mnc','.nii.gz'])
    randStr  = lambda: 'str_' + ''.join(rnd.choice(string.digits+string.ascii_letters) for _ in range(nd))

    # A function for generating a number type parameter input
    # p is a dictionary object corresponding to a parameter description in the json descriptor
    # E.g. if p had no constraints from the json descriptor, the output would be a float in [defaultMin,defaultMax]
    #      if p had "integer": true, "minimum": 7, "maximum": 9, the output would be an int in [7,9]
    def randNum(p):
      param_id, defaultMin, defaultMax = p['id'], -50, 50
      isInt = self.safeGet( param_id , 'integer' ) # Check if the input parameter should be an int
      roundTowardsZero = lambda x: int(math.copysign(1,x) * int(abs(x)))
      # Assign random values to min and max, unless they have been specified
      minv, maxv = self.safeGet( param_id , 'minimum' ), self.safeGet( param_id , 'maximum' )
      if minv is None and maxv is None: minv, maxv = defaultMin, defaultMax
      elif minv is None and not (maxv is None): minv = maxv + defaultMin
      elif not (minv is None) and maxv is None: maxv = minv + defaultMax
      # Coerce the min/max to the proper number type
      if isInt: minv, maxv = roundTowardsZero(minv), roundTowardsZero(maxv)
      else: minv, maxv = float(minv), float(maxv)
      # Apply exclusive boundary constraints, if any
      if self.safeGet( param_id , 'exclusive-minimum' ): minv += (1 if isInt else 0.0001)
      if self.safeGet( param_id , 'exclusive-maximum' ): maxv -= (1 if isInt else 0.0001)
      # Returns a random int or a random float, depending on the type of p
      return (rnd.randint(minv,maxv) if isInt else round(rnd.uniform(minv,maxv),nd) )

    # Generate a random parameter value based on the input type (where prm \in self.inputs)
    def paramSingle(prm):
      if self.safeGet(prm['id'],'value-choices'):   return rnd.choice( self.safeGet(prm['id'],'value-choices') )
      if prm['type']=='String': return randStr()
      if prm['type']=='Number': return randNum(prm)
      if prm['type']=='Flag':   return rnd.choice(['true','false'])
      if prm['type']=='File':   return randFile()

    # For this function, given prm (a parameter description), a parameter value is generated
    # If prm is a list, a sequence of outputs is generated; otherwise, a single value is returned
    def makeParam(prm):
      mn = self.safeGet(prm['id'], 'min-list-entries') or 2
      mx = self.safeGet(prm['id'], 'max-list-entries') or nl
      isList = self.safeGet(prm['id'], 'list') or False
      return ' '.join(str(paramSingle(prm)) for _ in range(rnd.randint(mn,mx))) if isList else paramSingle(prm)

    # Returns a list of the ids of parameters that disable the input parameter
    def disablersOf( inParam ):
      return [disabler[0] for disabler in [disabler for disabler in [(prm['id'], self.safeGet(prm['id'],'disables-inputs') or []) for prm in self.inputs] if inParam['id'] in disabler[1]]]

    # Returns the list of mutually requiring parameters of the target
    def mutReqs( targetParam ):
      return [self.byId( mutualReq[0] ) for mutualReq in [possibleMutReq for possibleMutReq in [(reqOfTarg, self.reqsOf(reqOfTarg)) for reqOfTarg in self.reqsOf(targetParam['id'])] if targetParam['id'] in possibleMutReq[1]]]

    # Returns whether targ (an input parameter) has a value or is allowed to have one
    def isOrCanBeFilled( targ ):
      # If it is already filled in, report so
      if targ['id'] in list(self.in_dict.keys()): return True
      # If a disabler or a disabled target is already active, it cannot be filled
      for d in disablersOf( targ ) + (self.safeGet( targ['id'], 'disables-inputs' ) or []):
        if d in list(self.in_dict.keys()): return False
      # If at least one non-mutual requirement has not been met, it cannot be filled
      for r in self.reqsOf(targ['id']):
        if not r in self.in_dict: # If a requirement is not present
          if not targ['id'] in self.reqsOf(r): # and it is not mutually required
            return False
      # If it is in a mutex group with one target already chosen, it cannot be filled
      g = self.assocGrp( targ['id'] ) # Get the group that the target belongs to, if any
      if (not g is None) and self.safeGrpGet(g['id'],'mutually-exclusive'):
        if len([x for x in g['members'] if x in list(self.in_dict.keys())]) > 0: return False
      return True

    # Handle the mutual requirement case by breadth first search in the graph of mutual requirements.
    # Essentially a graph is built, starting from targ (the target input parameter), where nodes are
    # input parameters and edges are a mutual requirement relation between nodes. BFS is used to check
    # every node, to see if can be (or has been) given a value. If all the nodes are permitted to be
    # given values, then they are all added at once; if even one of them cannot be given a value (e.g.
    # it has an active disabler) then none of the graph members can be added and so we just return false.
    #
    # Input: an input parameter from which to start building the graph
    # Output:
    #   Returns False if at least one of the mutual requirements cannot be met
    #   Returns a list of params to fill if all of them can be met (or [targ.id] if it has no mutReqs)
    def checkMutualRequirements(targ):
      checked, toCheck = [], [targ]
      while len(toCheck) > 0:
        current = toCheck.pop()
        checked.append( current )
        if not isOrCanBeFilled( current ): return False
        for mutreq in mutReqs( current ):
          if not mutreq['id'] in [c['id'] for c in checked]: toCheck.append(mutreq)
      return checked

    ## Start actual dictionary filling part ##
    # Clear the dictionary
    self.in_dict = {}
    # Fill in the required parameters
    for reqp in [r for r in self.inputs if not r.get('optional')]:
      print(reqp['id'])
      self.in_dict[reqp['id']] = makeParam(reqp)
    # Fill in a random choice for each one-is-required group
    for grp in [g for g in self.groups if self.safeGrpGet(g['id'],'one-is-required')]:
      while True: # Loop to choose an allowed value, in case a previous choice disabled that one
        choice = self.byId( rnd.choice( grp['members'] ) ) # Pick a random parameter
        res = checkMutualRequirements( choice ) # see if it and its mutual requirements can be filled
        if res==False: continue # Try again if the chosen group member is not permissable
        for r in res: self.in_dict[r['id']] = makeParam(r)
        break # If we were allowed to add a parameter, we can stop
    # Choose a random number of times to try to fill optional inputs
    opts = [p for p in self.inputs if self.safeGet(p['id'],'') in [None,True]]
    # Loop a random number of times, each time attempting to fill a random parameter
    for _ in range(rnd.randint( int(len(opts) / 2 + 1), len(opts) * 2)):
      targ = rnd.choice( opts ) # Choose an optional output
      # If it is already filled in, continue
      if targ['id'] in list(self.in_dict.keys()): continue
      # If it is a prohibited option, continue (isFilled case handled above)
      if not isOrCanBeFilled(targ): continue
      # Now we handle the mutual requirements case. This is a little more complex because a mutual requirement
      # of targ can have its own mutual requirements, ad nauseam. We need to look at all of them recursively and either
      # fill all of them in (i.e. at once) or none of them (e.g. if one of them is disabled by some other param).
      result = checkMutualRequirements( targ )
      # Leave if the mutreqs cannot be satisfied
      if result == False: continue
      # Fill in the target(s) otherwise
      for r in result: self.in_dict[r['id']] = makeParam(r)

  # Function to generate random parameter values
  # This fills the in_dict with random values, validates the input, and generates the appropriate command line
  def generateRandomParams(self,n):

    '''
    The generateRandomParams method fills the in_dict field with randomly generated values following the schema.
    It then generates command line strings based on these values (more than 1 if -n was given).
    '''

    self.cmdLine = []
    for i in range(0,n):
      # Set in_dict with random values
      self._randomFillInDict()
      # Look at generated input, if debugging
      if self.debug: print( "Input: " + str( self.in_dict ) )
      # Check results (as much as possible)
      try: self._validateDict()
      # If an error occurs, print out the problems already encountered before blowing up
      except Exception: # Avoid catching BaseExceptions like SystemExit
        sys.stderr.write("An error occurred in validation\nPreviously saved issues\n")
        for err in self.errs: sys.stderr.write("\t" + str(err) + "\n")
        raise # Pass on (throw) the caught exception
      # Add new command line
      self.cmdLine.append( self._generateCmdLineFromInDict() )

  # Read in parameter input file or string
  def readInput(self, infile):

    '''
    The readInput method sets the in_dict field of the executor object, based on a fixed input.
    It then generates a command line based on the input.

    infile: either the inputs in a file or the command-line string (from -s).
    stringInput: a boolean as to whether the method has been given a string or a file.
    '''

    # Quick check that the descriptor has already been read in
    assert self.desc_dict != None
    with open(infile, 'r') as inparams:
      self.in_dict = json.loads(inparams.read())
    # Input dictionary
    if self.debug: print( "Input: " + str( self.in_dict ) )
    # Fix special flag case: flags given the false value are treated as non-existent
    toRm = []
    for inprm in self.in_dict:
      if str(self.in_dict[inprm]).lower() == 'false' and self.byId(inprm)['type'] == 'Flag':
        toRm.append(inprm)
      elif self.byId(inprm)['type'] == 'Flag' and self.in_dict[inprm] == True:
        self.in_dict[inprm] = "true" # Fix json inputs using bools instead of strings
    for r in toRm: del self.in_dict[r]
    # Add default values for required parameters, if no value has been given
    for input in [s for s in self.inputs if not s.get("default-value") is None and not s.get("optional")]:
      if self.in_dict.get( input['id'] ) is None:
        df = input.get("default-value")
        self.in_dict[ input['id'] ] = df if not input['type'] == 'Flag' else str(df)
    # Check results (as much as possible)
    try: pass #self._validateDict()
    except Exception: # Avoid catching BaseExceptions like SystemExit
      sys.stderr.write("An error occurred in validation\nPreviously saved issues\n")
      for err in self.errs: sys.stderr.write("\t" + str(err) + "\n") # Write any errors we found
      raise # Raise the exception that caused failure
    # Build and save output command line (as a single-entry list)
    self.cmdLine = [ self._generateCmdLineFromInDict() ]

  # Private method to replace the keys in template by input and output
  # values. Input and output values are looked up in self.in_dict and
  # self.out_dict
  # * if useFlags is true, keys will be replaced by flag+flag-separator+value
  # * if unfoundKeys is "remove", unfound keys will be replaced by ""
  # * if unfoundKeys is "clear" then the template is cleared if it has unfound keys (useful for configuration files)
  # * before being substituted, the keys will be stripped from all the strings in strippedExtensions 
  def _replaceKeysInTemplate(self,template,useFlags=False,unfoundKeys="remove",strippedExtensions=[]):
      # Concatenate input and output dictionaries
      in_out_dict = dict(self.in_dict)
      in_out_dict.update(self.out_dict)
      # Go through all the keys
      for paramId in [x['id'] for x in self.inputs+self.outputs]:
        clk = self.safeGet(paramId,'value-key')
        if clk is None: continue          
        if paramId in list(in_out_dict.keys()): # param has a value
          val = in_out_dict[paramId]
          if type(val) is list:
            s_val = ""
            for x in val:
              s_val += x+" "
            val = s_val
          else:
              val = str(val)
          # Add flags and separator if necessary
          if useFlags:
            flag   = self.safeGet(paramId,'command-line-flag') or ''
            sep    = self.safeGet(paramId,'command-line-flag-separator') or ' '
            val = flag + sep + val  
            if self.safeGet(paramId,'type')=='Flag': # special case for flag-type inputs
              val = '' if val.lower()=='false' else flag
          # Remove file extensions from input value
          for extension in strippedExtensions:
            val = val.replace(extension,"")
          template = template.replace(clk,val)
        else: # param has no value
          if unfoundKeys == "remove":
            template = template.replace(clk, '')
          elif unfoundKeys == "clear":
            if clk in template: return ""
      return template
     
  # Private method to generate output file names.
  # Output file names will be put in self.out_dict.
  def _generateOutputFileNames(self):
    if not hasattr(self,'out_dict'):
      self.out_dict = {} # a dictionary that will contain the output file names
    for outputId in [x['id'] for x in self.outputs]:
      # Initialize file name with path template or existing value
      outputFileName = self.out_dict[outputId] if outputId in list(self.out_dict.keys()) else self.safeGet(outputId, 'path-template')
      strippedExtensions = self.safeGet(outputId, "path-template-stripped-extensions") or []
      # We keep the unfound keys because they will be substituted in a second call to the method in case they are output keys
      outputFileName = self._replaceKeysInTemplate(outputFileName,False,"keep",strippedExtensions)
      if self.safeGet(outputId, 'uses-absolute-path'):
        outputFileName = os.path.abspath(outputFileName)
      self.out_dict[outputId] = outputFileName

  # Private method to write configuration files
  # Configuration files are output files that have a file-template
  def _writeConfigurationFiles(self):
    for outputId in [x['id'] for x in self.outputs]:
      fileTemplate = self.safeGet(outputId,'file-template')
      if fileTemplate is None: continue # this is not a configuration file
      strippedExtensions = self.safeGet(outputId, "path-template-stripped-extensions") or []
      # We substitute the keys line by line so that we can clear the lines that have keys with no value (undefined optional params)
      newTemplate = []
      for line in fileTemplate:
        newTemplate.append(self._replaceKeysInTemplate(line,False,"clear",strippedExtensions))
      template="\n".join(newTemplate)
      # Write the configuration file
      fileName = self.out_dict[outputId]
      file = open(fileName,'w')
      file.write(template)
      file.close()
      
  # Private method to build the actual command line by substitution, using the input data
  def _generateCmdLineFromInDict(self):
    # Genrate output file names
    self._generateOutputFileNames()
    self._generateOutputFileNames() # it is required to call the method twice in case path templates contain output keys
    # Write configuration files
    self._writeConfigurationFiles()
    # Get the command line template
    template = self.desc_dict['command-line']
    # Substitute every given value into the template (incl. flags, flag-seps, ...)
    template = self._replaceKeysInTemplate(template,True,"remove")
    # Return substituted command line
    return template

  # Print the command line result
  def printCmdLine(self):
    print("Generated Command"+('s' if len(self.cmdLine)>1 else '')+':')
    for cmd in self.cmdLine: print(cmd)

  # Private method for validating input parameters
  def _validateDict(self):
    # Holder for errors
    self.errs = []
    # Return whether s is a proper number; if intCheck is true, also check if it is an int
    def isNumber(s,intCheck = False):
      try: int(s) if intCheck else float(s); return True
      except ValueError: return False
    # Check individual inputs
    for key in self.in_dict:
      isList = self.safeGet(key,"list")
      # Get current value and schema descriptor properties
      val, targ = self.in_dict[key], self.byId( key )
      # A little closure helper to check if input values satisy the json descriptor's constraints
      # Checks whether 'value' is appropriate for parameter 'input' by running 'isGood' on it
      # If the input parameter is bad, it adds 'msg' to the list of errors
      def check(keyname, isGood, msg, value): # Checks input values
        # No need to check constraints if they were not specified
        dontCheck = ((not keyname in list(targ.keys())) or (targ[keyname]==False))
        # Keyname = None is a flag to check the type
        if (not keyname is None) and dontCheck: return
        # The input function is used to check whether the input parameter is acceptable
        if not isGood(value, keyname):
          self.errs.append(key + ' (' + str(value) + ') ' + msg)
      # The id exists as an input and is not duplicate
      if len([ k for k in self.inputs if k['id']==key ]) != 1: self.errs.append(key+' is an invalid id')
      # Types are correct
      if targ["type"] == "Number":
        # Number type and constraints are not violated (note the lambda safety checks)
        for v in (str(val).split() if isList else [val]):
          check('minimum', lambda x,y: float(x) >= targ[y], "violates minimum value",v)
          check('exclusive-minimum', lambda x,y: float(x) > targ['minimum'], "violates exclusive min value",v)
          check('maximum', lambda x,y: float(x) <= targ[y], "violates maximum value",v)
          check('exclusive-maximum', lambda x,y: float(x) < targ['maximum'], "violates exclusive max value",v)
          check('integer', lambda x,y: isNumber(x,True), "violates integer requirement",v)
          check(None, lambda x,y: isNumber(x), "is not a number",v)
      elif self.safeGet(targ['id'],'value-choices'):
        # Value is in the list of allowed values
        check('value-choices', lambda x,y: x in targ[y], "is not a valid enum choice",val)
      elif targ["type"] == "Flag":
        # Should be 'true' or 'false' when lower-cased (based on our transformations of the input)
        check(None, lambda x,y: x.lower() in ["true","false"], "is not a valid flag value",val)
      elif targ["type"] == "File":
        # Check path-type (absolute vs relative)
        if not self.forcePathType:
          for ftarg in (str(val).split() if isList else [val]):
            check('uses-absolute-path', lambda x,y: os.path.isabs(x), "is not an absolute path",ftarg)
        else:
          # Replace incorrectly specified paths if desired
          replacementFiles = []
          launchDir = self.launchDir if (not self.launchDir is None) else os.getcwd()
          for ftarg in (str(val).split() if isList else [val]):
            # Special case 1: launchdir is specified and we want to use absolute path
            # Note: in this case, the pwd is mounted as the launchdir; we do not attempt to move files if they will not be mounted, currently
            # That is, specified files that are not in the pwd or a subfolder will not be mounted to the container
            if targ.get('uses-absolute-path') == True and (not self.launchDir is None):
              relpath = os.path.relpath(ftarg, os.getcwd()) # relative path to target, from the pwd
              mountedAbsPath = os.path.join(launchDir,relpath) # absolute path in the container
              replacementFiles.append( os.path.abspath(mountedAbsPath) )
            # If the input uses-absolute-path, replace the path with its absolute version
            elif targ.get('uses-absolute-path') == True: replacementFiles.append( os.path.abspath(ftarg) )
          # Replace old val with the new one
          self.in_dict[ key ] = " ".join( replacementFiles )
      # List length constraints are satisfied
      if isList: check('min-list-entries', lambda x,y: len(x.split()) >= targ[y], "violates min size",val)
      if isList: check('max-list-entries', lambda x,y: len(x.split()) <= targ[y], "violates max size",val)
    # Required inputs are present
    for reqId in [v['id'] for v in self.inputs if not v.get('optional')]:
      if not reqId in list(self.in_dict.keys()):
        self.errs.append('Required input ' + str(reqId) + ' is not present')
    # Disables/requires is satisfied
    for givenVal in [v for v in self.inputs if v['id'] in list(self.in_dict.keys())]:
      for r in self.reqsOf(givenVal['id']): # Check that requirements are present
        if not r in list(self.in_dict.keys()):
          self.errs.append('Input '+str(givenVal['id'])+' is missing requirement '+str(r))
      for d in (givenVal['disables-inputs'] if 'disables-inputs' in list(givenVal.keys()) else []):
        if d in list(self.in_dict.keys()): # Check if a disabler is present
          self.errs.append('Input '+str(d)+' should be disabled by '+str(givenVal['id']))
    # Group one-is-required/mutex is ok
    for group,mbs in [(x,x["members"]) for x in self.groups]:
      # Check that the set of parameters in mutually exclusive groups have at most one member present
      if ("mutually-exclusive" in list(group.keys())) and group["mutually-exclusive"]:
        if len(set.intersection( set(mbs) , set(self.in_dict.keys()) )) > 1:
          self.errs.append('Group ' + str(group["id"]) + ' is supposed to be mutex')
      # Check that the set of parameters in one-is-required groups have at least one member present
      if ("one-is-required" in list(group.keys())) and group["one-is-required"]:
        if len(set.intersection( set(mbs) , set(self.in_dict.keys()) )) < 1:
          self.errs.append('Group ' + str(group["id"]) + ' requires one member to be present')
    # Fast-fail if there was a problem with the input parameters
    if len(self.errs) != 0:
      sys.stderr.write("Problems found with prospective input:\n")
      for err in self.errs: sys.stderr.write("\t" + err + "\n")
      sys.exit(1)



from setuptools import setup

VERSION = "0.5.5"
DEPS = [
         "simplejson",
         "jsonschema",
         "gitpython",
         "PyGithub",
       ]

setup(name="boutiques",
      version=VERSION,
      description="Schema for describing bash command-line tools",
      url="http://github.com/boutiques/boutiques",
      author="Tristan Glatard, Gregory Kiar",
      author_email="tristan.glatard@gmail.com, gkiar07@gmail.com",
      classifiers=[
                "Programming Language :: Python",
                "Programming Language :: Python :: 2",
                "Programming Language :: Python :: 3",
                "Programming Language :: Python :: 2.7",
                "Programming Language :: Python :: 3.5",
                "Programming Language :: Python :: 3.6",
                "Programming Language :: Python :: 3.7",
                "Programming Language :: Python :: Implementation :: PyPy",
                "License :: OSI Approved :: MIT License",
                "Topic :: Software Development :: Libraries :: Python Modules",
                "Operating System :: OS Independent",
                "Natural Language :: English"
                  ],
      license="MIT",
      packages=["boutiques"],
      include_package_data=True,
      test_suite="pytest",
      tests_require=["pytest"],
      setup_requires=DEPS,
      install_requires=DEPS,
      entry_points = {
        "console_scripts": [
            "bosh=boutiques.bosh:bosh",
        ]
      },
      data_files=[("schema", ["boutiques/schema/descriptor.schema.json"]),
                  ("bids-template", ["boutiques/bids-app-template/template.json"]),
                  ("neurolinks-template", ["boutiques/neurolinks-template/tool.json"]),
                  ("exgood", ["boutiques/schema/examples/good.json"]),
                  ("exinvalid", ["boutiques/schema/examples/invalid.json"]),
                  ("exbidsgood", ["boutiques/schema/examples/bids_good.json"]),
                  ("exbidsbad1", ["boutiques/schema/examples/bids_bad1.json"]),
                  ("exbidsbad2", ["boutiques/schema/examples/bids_bad2.json"]),
                  ("exbad", ["boutiques/schema/examples/bad.json"])],
      zip_safe=False)
